#ifndef _BSP_TREE_H
#define _BSP_TREE_H

/*
*****************************************************************************************
	Main Header for BSP Tree
	This is very 'inefficient'
	It will still build a tree quickly but the use or virtual functions can be avoided
	I only coded it this way because my philosphy is to:
		MAKE IT RIGHT  then
		MAKE IT FAST
	And im still in exploration phase of my 3d career
******************************************************************************************
*/

extern num_NON_backface_polys;
extern double current_speed;

class BSP_Tree;

class BSP_Node {
	friend BSP_Tree;
protected:
	bool isLeaf;
	BSP_Node *front, *back;
public:
	BSP_Node(bool is_a_Leaf): isLeaf(is_a_Leaf), front(0),back(0) {}

	virtual void draw() {}
	virtual double facing(const Camera &c) const {return 0;}
	virtual Plane get_partition() const {return Plane();};
 };

class BSP_Partition: public BSP_Node {
private:
	Plane partition;
public:
	BSP_Partition(Plane &p): BSP_Node(false), partition(p) {}
	virtual ~BSP_Partition() {}

	virtual Plane  get_partition() const  {return partition;}
	virtual double facing(const Camera &c) const {return partition.dist(c.pos);}
};

class BSP_Leaf: public BSP_Node {
private:
	Poly_List polygons;
	Vect3D center;
	float radius;
	void bound_sphere();  // EXPERIMENTAL as of 2000/7/14
public:
	BSP_Leaf(Poly_List &list): BSP_Node(true), polygons(list) {bound_sphere();}
	virtual ~BSP_Leaf() {}

	virtual void draw() {
		for(poly_Iter=polygons.begin(); poly_Iter!=polygons.end(); ++poly_Iter) {
			if((*poly_Iter).plane.dist(player.pos)>=0.0F) {
			  (*poly_Iter).draw();
				num_NON_backface_polys++;
			}
		}
	}
};
/*
===========
BSP_Leaf::bound_sphere
	Graphics Gems modified source
	to find smallest bounding sphere of a given set of points
	Uses entire polygon list to find the leaf's bounding sphere
===========
*/
void BSP_Leaf::bound_sphere() {
	Poly curr_poly;
	Point3D curr_point, delta;
	Point3D xmin, xmax, ymin, ymax, zmin, zmax, dia1, dia2;
	double xspan, yspan, zspan, maxspan;
	double old_to_p, old_to_p_sq, old_to_new;
	double rad_sq;
	register int i;

	xmin.setx(INT_MAX); ymin.sety(INT_MAX); zmin.setz(INT_MAX);
	xmax.setx(INT_MIN); ymax.sety(INT_MIN); zmax.setz(INT_MIN);
	
	for (poly_Iter=polygons.begin(); poly_Iter!=polygons.end(); ++poly_Iter) {
		curr_poly=*poly_Iter;
		for(i=0; i<curr_poly.num_vertexes; i++) {
			curr_point=curr_poly.vertex[i];
			if (curr_point.getx() < xmin.getx())
				xmin = curr_point;
			if (curr_point.getx() > xmax.getx())
				xmax = curr_point;
			if (curr_point.gety() < ymin.gety())
				ymin = curr_point;
			if (curr_point.gety() > ymax.gety())
				ymax = curr_point;
			if (curr_point.getz() < zmin.getz())
				zmin = curr_point;
			if (curr_point.getz() > zmax.getz())
				zmax = curr_point;
			xmin.insure_0(); xmax.insure_0();
			ymin.insure_0(); ymax.insure_0();		
			zmin.insure_0(); zmax.insure_0();
		}
	}
	delta = xmax-xmin;
	xspan = delta.dot(delta);
	delta = ymax-ymin;
	yspan = delta.dot(delta);
	delta = zmax-zmin;
	zspan = delta.dot(delta);

	dia1 = xmin; dia2 = xmax;
	maxspan = xspan;
	if (yspan>maxspan) {
		maxspan = yspan;
		dia1 = ymin; dia2 = ymax;
	}
	if (zspan>maxspan) {
		dia1 = zmin; dia2 = zmax;
	}

	center.setx((dia1.getx()+dia2.getx())/2.0);
	center.sety((dia1.gety()+dia2.gety())/2.0);
	center.setz((dia1.getz()+dia2.getz())/2.0);
	center.insure_0();
	delta = dia2-center;
	rad_sq = delta.dot(delta);
	radius = sqrt(rad_sq);

	for (poly_Iter=polygons.begin(); poly_Iter!=polygons.end(); ++poly_Iter) {
		curr_poly=*poly_Iter;
		for(i=0; i<curr_poly.num_vertexes; i++) {
			curr_point=curr_poly.vertex[i];
			delta = curr_point-center;
			old_to_p_sq = delta.dot(delta);
			if (old_to_p_sq > rad_sq)	{
				old_to_p = sqrt(old_to_p_sq);
				radius = (radius + old_to_p) / 2.0;
				rad_sq = radius*radius;
				old_to_new = old_to_p - radius;
				center.setx((radius*center.getx() + old_to_new*curr_point.getx()) / old_to_p);
				center.sety((radius*center.gety() + old_to_new*curr_point.gety()) / old_to_p);
				center.setz((radius*center.getz() + old_to_new*curr_point.getz()) / old_to_p);
				center.insure_0();
			}	
		}
	}
}
//******************************************************************************************

class BSP_Tree {
private:
	void build_tree(Poly_List &, BSP_Node *);
	void render_tree(BSP_Node *);
protected:
	int num_elements, num_leafs;
	BSP_Node *root;
public:
	BSP_Tree(): num_elements(0), num_leafs(0), root(0) {}
	~BSP_Tree() { }

	int size()  {return num_elements;}
	int leafs() {return num_leafs;}

	void build(Poly_List &);
	void render() {render_tree(root);}
};
//******************************************************************************************
void BSP_Tree::build(Poly_List &polygons) {
	root=new BSP_Partition(polygons.front().plane);
	polygons.front().used=true;
	build_tree(polygons, root);
}

void BSP_Tree::build_tree(Poly_List &polygons, BSP_Node *node) {
	Poly_List front_list, back_list;

	partition(polygons, node->get_partition(), front_list, back_list);

	if(back_list.empty()) {
		node->back=0;
	}
	else {
		if(convex(back_list)) {
			num_elements+=back_list.size();
			num_leafs++;
			node->back=new BSP_Leaf(back_list);
		}
		else {
			for(poly_Iter=back_list.begin(); poly_Iter!=back_list.end(); ++poly_Iter) {
				if(!(*poly_Iter).used)	break;
			}
			(*poly_Iter).used=true;
			node->back=new BSP_Partition((*poly_Iter).plane);
			build_tree(back_list, node->back);
		}
	}
	if(front_list.empty()) {
		node->front=0;
	}
	else {
		if(convex(front_list)) {
			num_elements+=front_list.size();
			num_leafs++;
			node->front=new BSP_Leaf(front_list);
		}
		else {
			for(poly_Iter=front_list.begin(); poly_Iter!=front_list.end(); ++poly_Iter) {
				if(!(*poly_Iter).used)	break;
			}
			(*poly_Iter).used=true;
			node->front=new BSP_Partition((*poly_Iter).plane);
			build_tree(front_list, node->front);
		}
	}
}

void BSP_Tree::render_tree(BSP_Node *node) {
	if(node) {
		if(node->isLeaf) {
			node->draw();
			return;
		}
		if(node->facing(player) >= 0.0F) {
			render_tree(node->back);
			render_tree(node->front);
		}
		else {
			render_tree(node->front);
			render_tree(node->back);
		}
	}
}
#endif
